#!/usr/bin/env python3
# Time-stamp: "2021-09-25 15:39:38 (ylab)"

import numpy as np
import torch
import mngs


def fdr_correction(pvals, alpha=0.05, method="indep"):
    # https://github.com/mne-tools/mne-python/blob/main/mne/stats/multi_comp.py
    """P-value correction with False Discovery Rate (FDR).

    Correction for multiple comparison using FDR :footcite:`GenoveseEtAl2002`.

    This covers Benjamini/Hochberg for independent or positively correlated and
    Benjamini/Yekutieli for general or negatively correlated tests.

    Parameters
    ----------
    pvals : array_like
        Set of p-values of the individual tests.
    alpha : float
        Error rate.
    method : 'indep' | 'negcorr'
        If 'indep' it implements Benjamini/Hochberg for independent or if
        'negcorr' it corresponds to Benjamini/Yekutieli.

    Returns
    -------
    reject : array, bool
        True if a hypothesis is rejected, False if not.
    pval_corrected : array
    alpha : float
        Error rate.
    method : 'indep' | 'negcorr'
        If 'indep' it implements Benjamini/Hochberg for independent or if
        'negcorr' it corresponds to Benjamini/Yekutieli.

    Returns
    -------
    reject : array, bool
        True if a hypothesis is rejected, False if not.
    pval_corrected : array
    alpha : float
        Error rate.
    method : 'indep' | 'negcorr'
        If 'indep' it implements Benjamini/Hochberg for independent or if
        'negcorr' it corresponds to Benjamini/Yekutieli.

    Returns
    -------
    reject : array, bool
        True if a hypothesis is rejected, False if not.
    pval_corrected : array
        P-values adjusted for multiple hypothesis testing to limit FDR.

    Returns
    -------
    reject : array, bool
        True if a hypothesis is rejected, False if not.
    pval_corrected : array
        P-values adjusted for multiple hypothesis testing to limit FDR.

    References
    ----------
    .. footbibliography::
    """
    pvals = np.asarray(pvals)
    shape_init = pvals.shape
    pvals = pvals.ravel()

    pvals_sortind = np.argsort(pvals)
    pvals_sorted = pvals[pvals_sortind]
    sortrevind = pvals_sortind.argsort()

    if method in ["i", "indep", "p", "poscorr"]:
        ecdffactor = _ecdf(pvals_sorted)
    elif method in ["n", "negcorr"]:
        cm = np.sum(1.0 / np.arange(1, len(pvals_sorted) + 1))
        ecdffactor = _ecdf(pvals_sorted) / cm
    else:
        raise ValueError("Method should be 'indep' and 'negcorr'")

    reject = pvals_sorted < (ecdffactor * alpha)
    if reject.any():
        rejectmax = max(np.nonzero(reject)[0])
    else:
        rejectmax = 0
    reject[:rejectmax] = True

    pvals_corrected_raw = pvals_sorted / ecdffactor
    pvals_corrected = np.minimum.accumulate(pvals_corrected_raw[::-1])[::-1]
    pvals_corrected[pvals_corrected > 1.0] = 1.0
    pvals_corrected = pvals_corrected[sortrevind].reshape(shape_init)
    reject = reject[sortrevind].reshape(shape_init)
    return reject, pvals_corrected


def _ecdf(x):
    # https://github.com/mne-tools/mne-python/blob/main/mne/stats/multi_comp.py
    """No frills empirical cdf used in fdrcorrection."""
    nobs = len(x)
    return np.arange(1, nobs + 1) / float(nobs)


def _ecdf_torch(x):
    """No frills empirical cdf used in fdrcorrection."""
    nobs = len(x)
    return torch.arange(1, nobs + 1) / float(nobs)


def fdr_correction_torch(pvals, alpha=0.05, method="indep"):
    """P-value correction with False Discovery Rate (FDR).

    Correction for multiple comparison using FDR :footcite:`GenoveseEtAl2002`.

    This covers Benjamini/Hochberg for independent or positively correlated and
    Benjamini/Yekutieli for general or negatively correlated tests.

    Parameters
    ----------
    pvals : array_like
        Set of p-values of the individual tests.
    alpha : float
        Error rate.
    method : 'indep' | 'negcorr'
        If 'indep' it implements Benjamini/Hochberg for independent or if
        'negcorr' it corresponds to Benjamini/Yekutieli.

    Returns
    -------
    reject : array, bool
        True if a hypothesis is rejected, False if not.
    pval_corrected : array
    alpha : float
        Error rate.
    method : 'indep' | 'negcorr'
        If 'indep' it implements Benjamini/Hochberg for independent or if
        'negcorr' it corresponds to Benjamini/Yekutieli.

    Returns
    -------
    reject : array, bool
        True if a hypothesis is rejected, False if not.
    pval_corrected : array
    alpha : float
        Error rate.
    method : 'indep' | 'negcorr'
        If 'indep' it implements Benjamini/Hochberg for independent or if
        'negcorr' it corresponds to Benjamini/Yekutieli.

    Returns
    -------
    reject : array, bool
        True if a hypothesis is rejected, False if not.
    pval_corrected : array
        P-values adjusted for multiple hypothesis testing to limit FDR.

    Returns
    -------
    reject : array, bool
        True if a hypothesis is rejected, False if not.
    pval_corrected : array
        P-values adjusted for multiple hypothesis testing to limit FDR.

    References
    ----------
    .. footbibliography::
    """

    shape_init = pvals.shape
    pvals = pvals.ravel()

    pvals_sortind = torch.argsort(pvals)
    pvals_sorted = pvals[pvals_sortind]
    sortrevind = pvals_sortind.argsort()

    if method in ["i", "indep", "p", "poscorr"]:
        ecdffactor = _ecdf_torch(pvals_sorted)
    elif method in ["n", "negcorr"]:
        cm = torch.sum(1.0 / torch.arange(1, len(pvals_sorted) + 1))
        ecdffactor = _ecdf_torch(pvals_sorted) / cm
    else:
        raise ValueError("Method should be 'indep' and 'negcorr'")

    ecdffactor = ecdffactor.type_as(pvals_sorted)

    reject = pvals_sorted < (ecdffactor * alpha)

    if reject.any():
        rejectmax = max(torch.nonzero(reject)[0])
    else:
        rejectmax = 0
    reject[:rejectmax] = True

    pvals_corrected_raw = pvals_sorted / ecdffactor
    pvals_corrected_raw_flipped = pvals_corrected_raw.flip(0)
    pvals_cum_min = torch.zeros_like(pvals_corrected_raw_flipped)
    for ii in range(len(pvals_cum_min)):
        pvals_cum_min[ii] = (pvals_corrected_raw_flipped[0 : int(ii) + 1]).min()
    pvals_corrected = pvals_cum_min.flip(0)

    pvals_corrected[pvals_corrected > 1.0] = 1.0
    pvals_corrected = pvals_corrected[sortrevind].reshape(shape_init)
    reject = reject[sortrevind].reshape(shape_init)
    return reject, pvals_corrected


if __name__ == "__main__":
    pvals = [0.02, 0.03, 0.05]
    pvals_torch = torch.tensor(np.array([0.02, 0.03, 0.05]))

    reject, pvals_corrected = fdr_correction(pvals, alpha=0.05, method="indep")

    reject_torch, pvals_corrected_torch = fdr_correction_torch(
        pvals, alpha=0.05, method="indep"
    )

    arr = pvals_corrected.astype(float)
    tor = pvals_corrected_torch.numpy().astype(float)
    print(mngs.general.isclose(arr, tor))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-03-30 08:24:55 (ywatanabe)"
import os


def optuna_study(lpath, value_str, sort=False):
    """
    Loads an Optuna study and generates various visualizations for each target metric.

    Parameters:
    - lpath (str): Path to the Optuna study database.
    - value_str (str): The name of the column to be used as the optimization target.

    Returns:
    - None
    """
    import matplotlib

    matplotlib.use("Agg")
    import matplotlib.pyplot as plt
    import mngs
    import optuna
    import pandas as pd

    plt, CC = mngs.plt.configure_mpl(plt, fig_scale=3)

    lpath = lpath.replace("./", "/")

    study = optuna.load_study(study_name=None, storage=lpath)

    sdir = lpath.replace("sqlite:///", "./").replace(".db", "/")

    # To get the best trial:
    best_trial = study.best_trial
    print(f"Best trial number: {best_trial.number}")
    print(f"Best trial value: {best_trial.value}")
    print(f"Best trial parameters: {best_trial.params}")
    print(f"Best trial user attributes: {best_trial.user_attrs}")

    # Merge the user attributes into the study history DataFrame
    study_history = study.trials_dataframe().rename(
        columns={"value": value_str}
    )

    if sort:
        ascending = "MINIMIZE" in str(study.directions[0])  # [REVISED]
        study_history = study_history.sort_values(
            [value_str], ascending=ascending
        )

    # Add user attributes to the study history DataFrame
    attrs_df = []
    for trial in study.trials:
        user_attrs = trial.user_attrs
        user_attrs = {k: v for k, v in user_attrs.items()}
        attrs_df.append({"number": trial.number, **user_attrs})
    attrs_df = pd.DataFrame(attrs_df).set_index("number")

    # Updates study history
    study_history = study_history.merge(
        attrs_df, left_index=True, right_index=True, how="left"
    ).set_index("number")
    try:
        study_history = mngs.gen.mv_col(study_history, "SDIR", 1)
        study_history["SDIR"] = study_history["SDIR"].apply(
            lambda x: str(x).replace("RUNNING", "FINISHED")
        )
        best_trial_dir = study_history["SDIR"].iloc[0]
        mngs.gen.symlink(best_trial_dir, sdir + "best_trial", force=True)
    except Exception as e:
        print(e)
    mngs.io.save(study_history, sdir + "study_history.csv")
    print(study_history)

    # To visualize the optimization history:
    fig = optuna.visualization.plot_optimization_history(
        study, target_name=value_str
    )
    mngs.io.save(fig, sdir + "optimization_history.png")
    mngs.io.save(fig, sdir + "optimization_history.html")
    plt.close()

    # To visualize the parameter importances:
    fig = optuna.visualization.plot_param_importances(
        study, target_name=value_str
    )
    mngs.io.save(fig, sdir + "param_importances.png")
    mngs.io.save(fig, sdir + "param_importances.html")
    plt.close()

    # To visualize the slice of the study:
    fig = optuna.visualization.plot_slice(study, target_name=value_str)
    mngs.io.save(fig, sdir + "slice.png")
    mngs.io.save(fig, sdir + "slice.html")
    plt.close()

    # To visualize the contour plot of the study:
    fig = optuna.visualization.plot_contour(study, target_name=value_str)
    mngs.io.save(fig, sdir + "contour.png")
    mngs.io.save(fig, sdir + "contour.html")
    plt.close()

    # To visualize the parallel coordinate plot of the study:
    fig = optuna.visualization.plot_parallel_coordinate(
        study, target_name=value_str
    )
    mngs.io.save(fig, sdir + "parallel_coordinate.png")
    mngs.io.save(fig, sdir + "parallel_coordinate.html")
    plt.close()


if __name__ == "__main__":
    mngs.plt.configure_mpl(plt, fig_scale=3)
    lpath = "sqlite:///scripts/ml/clf/sub_conv_transformer_optuna/optuna_studies/optuna_study_v001.db"
    lpath = "sqlite:///scripts/ml/clf/rocket_optuna/optuna_studies/optuna_study_v001.db"
    optuna_study(lpath, "Validation bACC")
    # scripts/ml/clf/sub_conv_transformer/optuna_studies/optuna_study_v032

    lpath = "sqlite:///scripts/ml/clf/sub_conv_transformer_optuna/optuna_studies/optuna_study_v020.db"
    mngs.ml.plt.optuna_study(lpath, "val_loss", sort=True)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-05-14 00:53:51 (ywatanabe)"

import matplotlib.pyplot as plt
import mngs
import numpy as np
import seaborn as sns
import umap.umap_ as umap_orig
from natsort import natsorted
from sklearn.preprocessing import LabelEncoder


def umap(
    data_all,
    labels_all,
    axes_titles=None,
    supervised=False,
    title="UMAP Clustering",
    alpha=0.1,
    s=3,
    use_independent_legend=False,
    add_super_imposed=False,
    palette="viridis",
):

    assert len(data_all) == len(labels_all)

    if isinstance(data_all, list):
        data_all = list(data_all)
        labels_all = list(labels_all)

    le = LabelEncoder()
    # le.fit(np.hstack(labels_all))
    le.fit(natsorted(np.hstack(labels_all)))
    labels_all = [le.transform(labels) for labels in labels_all]

    umap_model = umap_orig.UMAP(random_state=42)

    if supervised:
        _umap = umap_model.fit(data_all[0], y=labels_all[0])
        title = f"(Supervised) {title}"
    else:
        _umap = umap_model.fit(data_all[0])
        title = f"(Unsupervised) {title}"

    ncols = len(data_all) + 1 if add_super_imposed else len(data_all)
    share = True if ncols > 1 else False
    fig, axes = plt.subplots(ncols=ncols, sharex=share, sharey=share)
    fig.suptitle(title)
    fig.supxlabel("UMAP 1")
    fig.supylabel("UMAP 2")

    for ii, (data, labels) in enumerate(zip(data_all, labels_all)):
        embedding = _umap.transform(data)

        if ncols == 1:
            ax = axes
        else:
            ax = axes[ii + 1] if add_super_imposed else axes[ii]

        sns.scatterplot(
            x=embedding[:, 0],
            y=embedding[:, 1],
            hue=le.inverse_transform(labels),
            ax=ax,
            palette=palette,
            s=s,
            alpha=alpha,
        )

        ax.set_box_aspect(1)

        if axes_titles is not None:
            ax.set_title(axes_titles[ii])

        if not use_independent_legend:
            ax.legend(loc="upper left")

        if add_super_imposed:
            axes[0].set_title("Superimposed")
            axes[0].set_aspect("equal")
            sns.scatterplot(
                x=embedding[:, 0],
                y=embedding[:, 1],
                hue=le.inverse_transform(labels),
                ax=axes[0],
                palette=palette,
                legend="full" if ii == 0 else False,
                s=s,
                alpha=alpha,
            )

    if not use_independent_legend:
        return fig, None, _umap

    elif use_independent_legend:
        legend_figs = []
        for i, ax in enumerate(axes):
            legend = ax.get_legend()
            if legend:
                legend_fig = plt.figure(figsize=(3, 2))
                new_legend = legend_fig.gca().legend(
                    handles=legend.legendHandles,
                    labels=legend.texts,
                    loc="center",
                )
                legend_fig.canvas.draw()
                legend_filename = f"legend_{i}.png"
                legend_fig.savefig(legend_filename, bbox_inches="tight")
                legend_figs.append(legend_fig)
                plt.close(legend_fig)

        for ax in axes:
            ax.legend_ = None

        return fig, legend_figs, _umap



#!/usr/bin/env python3


def extend(ax, x_ratio=1.0, y_ratio=1.0):
    ## Original coordinates
    bbox = ax.get_position()
    left_orig = bbox.x0
    bottom_orig = bbox.y0
    width_orig = bbox.x1 - bbox.x0
    height_orig = bbox.y1 - bbox.y0
    g_orig = (left_orig + width_orig / 2.0, bottom_orig + height_orig / 2.0)

    ## Target coordinates
    g_tgt = g_orig
    width_tgt = width_orig * x_ratio
    height_tgt = height_orig * y_ratio
    left_tgt = g_tgt[0] - width_tgt / 2
    bottom_tgt = g_tgt[1] - height_tgt / 2

    ax.set_position(
        [
            left_tgt,
            bottom_tgt,
            width_tgt,
            height_tgt,
        ]
    )
    return ax


if __name__ == "__main__":
    import matplotlib.pyplot as plt

    fig, axes = plt.subplots(1, 2)
    ax = axes[1]
    ax = extend(ax, 0.75, 1.1)
    fig.show()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2022-12-09 13:38:11 (ywatanabe)"


def set_size(ax, w, h):
    """w, h: width, height in inches"""
    # if not ax: ax=plt.gca()
    l = ax.figure.subplotpars.left
    r = ax.figure.subplotpars.right
    t = ax.figure.subplotpars.top
    b = ax.figure.subplotpars.bottom
    figw = float(w) / (r - l)
    figh = float(h) / (t - b)
    ax.figure.set_size_inches(figw, figh)
    return ax



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-06-04 08:32:27 (ywatanabe)"
# /home/ywatanabe/proj/mngs/src/mngs/general/_wrap.py


import functools


def wrap(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-04-05 12:15:42 (ywatanabe)"

import torch as _torch
from mngs.general import torch_fn as _torch_fn


@_torch_fn
def z(x, dim=-1):
    return (x - x.mean(dim=dim, keepdim=True)) / x.std(dim=dim, keepdim=True)


@_torch_fn
def minmax(x, amp=1.0, dim=-1, fn="mean"):
    MM = x.max(dim=dim, keepdims=True)[0].abs()
    mm = x.min(dim=dim, keepdims=True)[0].abs()
    return amp * x / _torch.maximum(MM, mm)



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-04-17 09:34:43"
# Author: Yusuke Watanabe (ywata1989@gmail.com)

"""
This script does XYZ.
"""

import fnmatch
import os
import sys

import matplotlib.pyplot as plt

# Imports
import mngs


# Functions
def find_git_root():
    import git

    repo = git.Repo(".", search_parent_directories=True)
    return repo.working_tree_dir


def find_dir(root_dir, exp):
    return _find(root_dir, type="d", exp=exp)


def find_file(root_dir, exp):
    return _find(root_dir, type="f", exp=exp)


def _find(rootdir, type="f", exp=["*"]):
    """
    Mimicks the Unix find command.

    Example:
        # rootdir =
        # type = 'f'  # 'f' for files, 'd' for directories, None for both
        # exp = '*.txt'  # Pattern to match, or None to match all
        find('/path/to/search', "f", "*.txt")
    """
    if isinstance(exp, str):
        exp = [exp]

    matches = []
    for _exp in exp:
        for root, dirs, files in os.walk(rootdir):
            # Depending on the type, choose the list to iterate over
            if type == "f":  # Files only
                names = files
            elif type == "d":  # Directories only
                names = dirs
            else:  # All entries
                names = files + dirs

            for name in names:
                # Construct the full path
                path = os.path.join(root, name)

                # If an _exp is provided, use fnmatch to filter names
                if _exp and not fnmatch.fnmatch(name, _exp):
                    continue

                # If type is set, ensure the type matches
                if type == "f" and not os.path.isfile(path):
                    continue
                if type == "d" and not os.path.isdir(path):
                    continue

                exclude_keys = ["/lib/", "/env/", "/build/"]
                if not any(ek in path for ek in exclude_keys):
                    matches.append(path)

                # for ek in exclude_keys:
                #     if ek in path:
                #         path = None
                #         break

                # if path is not None:
                #     # Add the matching path to the results
                #     matches.append(path)

    return matches


if __name__ == "__main__":
    # Start
    CONFIG, sys.stdout, sys.stderr, plt, CC = mngs.gen.start(sys, plt)

    # (YOUR AWESOME CODE)

    # Close
    mngs.gen.close(CONFIG)

# EOF

"""
/ssh:ywatanabe@444:/home/ywatanabe/proj/entrance/mngs/path/_find.py
"""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-06-03 08:33:33 (ywatanabe)"#!/usr/bin/env python3

import warnings
from functools import wraps

import numpy as np
import pandas as pd
import torch
import xarray
from tqdm import tqdm


def try_device(x, device):
    if not isinstance(x, torch.Tensor):
        return x

    # When no conversion is necessary
    orig_device = x.device
    if orig_device == device:
        return x

    try:
        x_new = x.to(device)
        return x_new  # When successful

    except RuntimeError as e:
        if "cuda" in str(e).lower() and device == "cuda":
            warnings.warn(
                "CUDA memory insufficient, falling back to CPU.",
                UserWarning,
            )
            x_new = x.cpu()
            return x_new
        else:
            raise e


def unsqueeze_if(arr_or_tensor, ndim=2, axis=0):
    if isinstance(arr_or_tensor, (np.ndarray, torch.Tensor)):
        if arr_or_tensor.ndim != ndim:
            raise ValueError(f"Input must be a {ndim}-D array or tensor.")
        return torch.unsqueeze(arr_or_tensor, dim=axis)
    else:
        raise TypeError("Input must be a NumPy array or a PyTorch tensor.")


def squeeze_if(arr_or_tensor, ndim=3, axis=0):
    if isinstance(arr_or_tensor, (np.ndarray, torch.Tensor)):
        if arr_or_tensor.ndim != ndim:
            raise ValueError(f"Input must be a {ndim}-D array or tensor.")
        return torch.squeeze(arr_or_tensor, dim=axis)
    else:
        raise TypeError("Input must be a NumPy array or a PyTorch tensor.")


def is_torch(*args, **kwargs):
    return any(isinstance(arg, torch.Tensor) for arg in args) or any(
        isinstance(v, torch.Tensor) for v in kwargs.values()
    )


def is_cuda(*args, **kwargs):
    return any(
        [(isinstance(arg, torch.Tensor) and arg.is_cuda) for arg in args]
    ) or any(
        [(isinstance(v, torch.Tensor) and v.is_cuda) for v in kwargs.values()]
    )


def return_always(*args, **kwargs):
    return args, kwargs


def return_if(*args, **kwargs):
    if (len(args) > 0) and (len(kwargs) > 0):
        return args, kwargs
    elif (len(args) > 0) and (len(kwargs) == 0):
        return args
    elif (len(args) == 0) and (len(kwargs) > 0):
        return kwargs
    elif (len(args) == 0) and (len(kwargs) == 0):
        return None


def to_torch(*args, return_fn=return_if, **kwargs):
    def _conversion_warning(x, x_new):
        warnings.warn(
            f"Converted from {type(x).__name__} to {type(x_new).__name__} ({x_new.device}). "
            f"You might want to consider using {type(x_new).__name__} ({x_new.device}) as input for faster computation.",
            category=UserWarning,
        )

    def _to_torch(x, device=kwargs.get("device")):
        if device is None:
            device = "cuda" if torch.cuda.is_available() else "cpu"

        # Pandas
        if isinstance(x, (pd.Series, pd.DataFrame)):
            x_new = torch.tensor(x.to_numpy()).squeeze().float()
            x_new = try_device(x_new, device)
            if device == "cuda":
                _conversion_warning(x, x_new)
            return x_new

        # Numpy
        elif isinstance(x, (np.ndarray, list)):
            x_new = torch.tensor(x).float()
            x_new = try_device(x_new, device)
            if device == "cuda":
                _conversion_warning(x, x_new)
            return x_new

        # xarray
        elif isinstance(x, xarray.core.dataarray.DataArray):
            x_new = torch.tensor(np.array(x)).float()
            x_new = try_device(x_new, device)
            if device == "cuda":
                _conversion_warning(x, x_new)
            return x_new

        # Tuple
        elif isinstance(x, tuple):
            return [_to_torch(_x) for _x in x if _x is not None]

        else:
            x_new = x
            return x_new

    # def _to_torch(x, device=kwargs.get("device", "cuda")):
    #     # pandas
    #     if isinstance(x, (pd.Series, pd.DataFrame)):
    #         x = x.to_numpy()

    #     # numpy
    #     if isinstance(x, (np.ndarray, list)):
    #         x = torch.tensor(x)
    #         x_new = try_device(x, device)
    #         return x_new

    #     # Else
    #     else:
    #         # Directly return non-convertible types without warning
    #         return x

    c_args = [_to_torch(arg) for arg in args if arg is not None]
    c_kwargs = {k: _to_torch(v) for k, v in kwargs.items() if v is not None}

    # Handle renaming 'axis' to 'dim' for PyTorch compatibility
    if "axis" in c_kwargs:
        c_kwargs["dim"] = c_kwargs.pop("axis")

    return return_fn(*c_args, **c_kwargs)


# def to_torch(*args, return_fn=return_if, **kwargs):
#     def _to_torch(x, device=kwargs.get("device")):
#         if device is None:
#             device = "cuda" if torch.cuda.is_available() else "cpu"

#         if isinstance(x, (pd.Series, pd.DataFrame)):
#             x_new = torch.tensor(x.to_numpy()).squeeze().float().to(device)
#             warnings.warn(
#                 f"Converted from  {type(x).__name__} to {type(x_new).__name__} ({x_new.device}) If you prioritize calculation speed, please consider to input with torch.tensor instead of {type(x).__name__}",
#                 category=UserWarning,
#             )
#             return x_new

#         elif isinstance(x, (np.ndarray, list)):
#             x_new = torch.tensor(x).float().to(device)
#             warnings.warn(
#                 f"Converted from  {type(x).__name__} to {type(x_new).__name__} ({x_new.device})",
#                 category=UserWarning,
#             )
#             return x_new

#         else:
#             return x

#     c_args = [_to_torch(arg) for arg in args if arg is not None]
#     c_kwargs = {k: _to_torch(v) for k, v in kwargs.items() if v is not None}

#     if "axis" in c_kwargs:
#         c_kwargs["dim"] = c_kwargs.pop("axis")

#     return return_fn(*c_args, **c_kwargs)


def to_numpy(*args, return_fn=return_if, **kwargs):
    def _to_numpy(x):
        if isinstance(x, (pd.Series, pd.DataFrame)):
            return x.to_numpy().squeeze()
        elif isinstance(x, torch.Tensor):
            return x.detach().cpu().numpy()
        elif isinstance(x, list):
            return np.array(x)
        elif isinstance(x, tuple):
            return [_to_numpy(_x) for _x in x if _x is not None]
        else:
            return x

    c_args = [_to_numpy(arg) for arg in args if arg is not None]
    c_kwargs = {k: _to_numpy(v) for k, v in kwargs.items() if v is not None}

    if "dim" in c_kwargs:
        c_kwargs["axis"] = c_kwargs.pop("dim")

    return return_fn(*c_args, **c_kwargs)


def torch_fn(func):
    """
    Decorator to ensure torch calculation

    Example:
        import torch.nn.functional as F

        @torch_fn
        def torch_softmax(*args, **kwargs):
            return F.softmax(*args, **kwargs)

        def custom_print(x):
            print(type(x), x)

        x = [1, 2, 3]
        x_list = x
        x_tensor = torch.tensor(x).float()
        x_tensor_cuda = torch.tensor(x).float().cuda()
        x_array = np.array(x)
        x_df = pd.DataFrame({"col1": x})

        custom_print(torch_softmax(x_list, dim=-1))
        custom_print(torch_softmax(x_array, dim=-1))
        custom_print(torch_softmax(x_df, dim=-1))
        custom_print(torch_softmax(x_tensor, dim=-1))
        custom_print(torch_softmax(x_tensor_cuda, dim=-1))

        # <class 'numpy.ndarray'> [0.09003057 0.24472848 0.66524094]
        # <class 'numpy.ndarray'> [0.09003057 0.24472848 0.66524094]
        # <class 'numpy.ndarray'> [0.09003057 0.24472848 0.66524094]
        # <class 'torch.Tensor'> tensor([0.0900, 0.2447, 0.6652])
        # <class 'torch.Tensor'> tensor([0.0900, 0.2447, 0.6652], device='cuda:0')
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        # Buffers the input types
        is_torch_input = is_torch(*args, **kwargs)

        # Runs the func
        c_args, c_kwargs = to_torch(*args, return_fn=return_always, **kwargs)

        results = func(*c_args, **c_kwargs)

        # Reverts to the original data type
        if not is_torch_input:
            return to_numpy(results, return_fn=return_if)[0]

        elif is_torch_input:
            return results

    return wrapper


def numpy_fn(func):
    """
    Decorator to ensure numpy calculation

    Example:
        import scipy

        @numpy_fn
        def numpy_softmax(*args, **kwargs):
            return scipy.special.softmax(*args, **kwargs)

        def custom_print(x):
            print(type(x), x)

        x = [1, 2, 3]
        x_list = x
        x_tensor = torch.tensor(x).float()
        x_tensor_cuda = torch.tensor(x).float().cuda()
        x_array = np.array(x)
        x_df = pd.DataFrame({"col1": x})

        custom_print(numpy_softmax(x_list, axis=-1))
        custom_print(numpy_softmax(x_array, axis=-1))
        custom_print(numpy_softmax(x_df, axis=-1))
        custom_print(numpy_softmax(x_tensor, axis=-1))
        custom_print(numpy_softmax(x_tensor_cuda, axis=-1))

        # <class 'numpy.ndarray'> [0.09003057 0.24472847 0.66524096]
        # <class 'numpy.ndarray'> [0.09003057 0.24472847 0.66524096]
        # <class 'numpy.ndarray'> [0.09003057 0.24472847 0.66524096]
        # <class 'torch.Tensor'> tensor([0.0900, 0.2447, 0.6652])
        # <class 'torch.Tensor'> tensor([0.0900, 0.2447, 0.6652], device='cuda:0')
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        # Buffers the input types
        is_torch_input = is_torch(*args, **kwargs)
        is_cuda_input = is_cuda(*args, **kwargs)
        device = "cuda" if is_cuda_input else "cpu"

        # Runs the func
        c_args, c_kwargs = to_numpy(*args, return_fn=return_always, **kwargs)
        results = func(*c_args, **c_kwargs)

        # Reverts to the original data type
        if not is_torch_input:
            return results
        else:
            return to_torch(results, return_fn=return_if, device=device)[0][0]

    return wrapper


def batch_fn(func):
    """
    Decorator to apply batch processing to a function. This is useful when the input tensor is too large to process at once due to memory constraints.

    Args:
    - func (Callable): A function that processes a tensor.

    Returns:
    - Callable: A wrapped function that processes the tensor in batches.
    """

    @wraps(func)
    def wrapper(x, *args, **kwargs):
        # Get batch_size from kwargs with a default value of 4
        batch_size = int(kwargs.pop("batch_size", 4))

        if len(x) <= batch_size:
            return func(x, *args, **kwargs, batch_size=batch_size)

        n_batches = (len(x) + batch_size - 1) // batch_size
        results = []

        for i_batch in tqdm(range(n_batches)):
            start = i_batch * batch_size
            end = min((i_batch + 1) * batch_size, len(x))
            batch_result = func(
                x[start:end], *args, **kwargs, batch_size=batch_size
            )

            if isinstance(batch_result, torch.Tensor):
                batch_result = batch_result.cpu()
            elif isinstance(batch_result, tuple):
                batch_result = tuple(
                    vv.cpu() if isinstance(vv, torch.Tensor) else vv
                    for vv in batch_result
                )

            results.append(batch_result)

        # Check if the function returns a tuple of results or a single result
        if isinstance(results[0], tuple):
            # Handle multiple outputs
            n_vars = len(results[0])
            combined_results = [
                torch.vstack([res[i_var] for res in results])
                for i_var in range(n_vars)
            ]
            return tuple(combined_results)
        else:
            # Handle single output
            if isinstance(results[0], torch.Tensor):
                return torch.vstack(results)
            else:
                # If the single output is not a tensor, concatenate or combine them as needed
                return sum(results, [])

    return wrapper


# def batch_fn(batch_size=4):
#     def decorator(func):
#         @wraps(func)
#         def wrapper(x, *args, **kwargs):
#             if len(x) <= batch_size:
#                 return func(x, *args, **kwargs)

#             n_batches = (len(x) + batch_size - 1) // batch_size
#             results = []

#             for i_batch in tqdm(range(n_batches)):
#                 start = i_batch * batch_size
#                 end = min((i_batch + 1) * batch_size, len(x))
#                 batch_result = func(x[start:end], *args, **kwargs)

#                 if isinstance(batch_result, torch.Tensor):
#                     batch_result = batch_result.cpu()
#                 elif isinstance(batch_result, tuple):
#                     batch_result = tuple(
#                         vv.cpu() if isinstance(vv, torch.Tensor) else vv
#                         for vv in batch_result
#                     )

#                 results.append(batch_result)

#             # Check if the function returns a tuple of results or a single result
#             if isinstance(results[0], tuple):
#                 # Handle multiple outputs
#                 n_vars = len(results[0])
#                 combined_results = [
#                     torch.vstack([res[i_var] for res in results])
#                     for i_var in range(n_vars)
#                 ]
#                 return tuple(combined_results)
#             else:
#                 # Handle single output
#                 if isinstance(results[0], torch.Tensor):
#                     return torch.vstack(results)
#                 else:
#                     # If the single output is not a tensor, concatenate or combine them as needed
#                     return sum(results, [])

#         return wrapper

#     return decorator


if __name__ == "__main__":
    import scipy
    import torch.nn.functional as F

    @torch_fn
    def torch_softmax(*args, **kwargs):
        return F.softmax(*args, **kwargs)

    @numpy_fn
    def numpy_softmax(*args, **kwargs):
        return scipy.special.softmax(*args, **kwargs)

    def custom_print(x):
        print(type(x), x)

    # Test the decorator with different input types
    x = [1, 2, 3]
    x_list = x
    x_tensor = torch.tensor(x).float()
    x_tensor_cuda = torch.tensor(x).float().cuda()
    x_array = np.array(x)
    x_df = pd.DataFrame({"col1": x})

    custom_print(torch_softmax(x_list, dim=-1))
    custom_print(torch_softmax(x_array, dim=-1))
    custom_print(torch_softmax(x_df, dim=-1))
    custom_print(torch_softmax(x_tensor, dim=-1))
    custom_print(torch_softmax(x_tensor_cuda, dim=-1))
    # <class 'numpy.ndarray'> [0.09003057 0.24472848 0.66524094]
    # <class 'numpy.ndarray'> [0.09003057 0.24472848 0.66524094]
    # <class 'numpy.ndarray'> [0.09003057 0.24472848 0.66524094]
    # <class 'torch.Tensor'> tensor([0.0900, 0.2447, 0.6652])
    # <class 'torch.Tensor'> tensor([0.0900, 0.2447, 0.6652], device='cuda:0')

    custom_print(numpy_softmax(x_list, axis=-1))
    custom_print(numpy_softmax(x_array, axis=-1))
    custom_print(numpy_softmax(x_df, axis=-1))
    custom_print(numpy_softmax(x_tensor, axis=-1))
    custom_print(numpy_softmax(x_tensor_cuda, axis=-1))
    # <class 'numpy.ndarray'> [0.09003057 0.24472847 0.66524096]
    # <class 'numpy.ndarray'> [0.09003057 0.24472847 0.66524096]
    # <class 'numpy.ndarray'> [0.09003057 0.24472847 0.66524096]
    # <class 'torch.Tensor'> tensor([0.0900, 0.2447, 0.6652])
    # <class 'torch.Tensor'> tensor([0.0900, 0.2447, 0.6652], device='cuda:0')



#!/usr/bin/env python3

from ._conf_mat import conf_mat
from ._learning_curve import learning_curve
from ._optuna_study import optuna_study
from .aucs.pre_rec_auc import pre_rec_auc
from .aucs.roc_auc import roc_auc



#!/usr/bin/env python3

from mngs.dsp import wavelet
import numpy as np


def test_wavelet():
    signal = np.random.rand(1024)
    samp_rate = 512
    Nyq = int(samp_rate / 2)
    try:
        wavelet(signal, samp_rate, f_min=10, f_max=Nyq, plot=False)
    except:
        assert False


test_wavelet()

"""
spy -m pytest ./test/mngs/dsp/test_wavelet.py
"""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-06-06 08:13:23 (ywatanabe)"

import os
from datetime import datetime
from glob import glob
from time import sleep

import mngs


def format_diff_time(diff_time):
    # Get total seconds from the timedelta object
    total_seconds = int(diff_time.total_seconds())

    # Calculate hours, minutes and seconds
    hours = total_seconds // 3600
    minutes = (total_seconds % 3600) // 60
    seconds = total_seconds % 60

    # Format the time difference as a string
    diff_time_str = f"{hours:02d}:{minutes:02d}:{seconds:02d}"
    return diff_time_str


def close(CONFIG, message=":)", notify=True, verbose=True):
    try:
        CONFIG["END_TIME"] = datetime.now()
        CONFIG["SPENT_TIME"] = format_diff_time(
            CONFIG["END_TIME"] - CONFIG["START_TIME"]
        )
        if verbose:
            print(f"\nEND TIME: {CONFIG['END_TIME']}")
            print(f"\nSPENT TIME: {CONFIG['SPENT_TIME']}")

    except Exception as e:
        print(e)

    mngs.io.save(
        CONFIG, CONFIG["SDIR"] + "CONFIGS/CONFIG.pkl", verbose=verbose
    )
    mngs.io.save(CONFIG, CONFIG["SDIR"] + "CONFIGS/CONFIG.yaml", verbose=False)

    try:
        if CONFIG.get("DEBUG", False):
            message = f"[DEBUG]\n" + message
        sleep(1)
        if notify:
            mngs.gen.notify(
                message=message,
                ID=CONFIG["ID"],
                log_paths=glob(CONFIG["SDIR"] + "logs/*.log"),
                verbose=verbose,
            )
    except Exception as e:
        print(e)

    # RUNNING to FINISHED
    finish(CONFIG["SDIR"])


def finish(src_dir, remove_src_dir=True):
    dest_dir = src_dir.replace("RUNNING/", "FINISHED/")
    os.makedirs(dest_dir, exist_ok=True)
    try:
        os.rename(src_dir, dest_dir)
        mngs.gen.print_block(
            f"Congratulations! The script completed.\n\n{dest_dir}",
            c="yellow",
        )
        if remove_src_dir:
            mngs.sh(f"rm {src_dir} -rf", verbose=False)
    except Exception as e:
        print(e)


if __name__ == "__main__":
    import sys

    import matplotlib.pyplot as plt
    import mngs
    from icecream import ic

    CONFIG, sys.stdout, sys.stderr, plt, CC = mngs.gen.start(
        sys, plt, verbose=False
    )

    ic("aaa")
    ic("bbb")
    ic("ccc")

    close(CONFIG)



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-04-11 08:26:19 (ywatanabe)"

import torch
import torch.nn.functional as F
from mngs.general import torch_fn


def _zero_pad_1d(x, target_length):
    padding_needed = target_length - len(x)
    padding_left = padding_needed // 2
    padding_right = padding_needed - padding_left
    return F.pad(x, (padding_left, padding_right), "constant", 0)


def zero_pad(xs, dim=0):
    max_len = max([len(x) for x in xs])
    return torch.stack([_zero_pad_1d(x, max_len) for x in xs], dim=dim)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-03-02 09:52:28 (ywatanabe)"

from warnings import warn

import numpy as np
import pandas as pd
import torch
from sklearn.preprocessing import LabelEncoder as SklearnLabelEncoder


class LabelEncoder(SklearnLabelEncoder):
    """
    An extension of the sklearn.preprocessing.LabelEncoder that supports incremental learning.
    This means it can handle new classes without forgetting the old ones.

    Attributes:
        classes_ (np.ndarray): Holds the label for each class.

    Example usage:
        encoder = IncrementalLabelEncoder()
        encoder.fit(np.array(["apple", "banana"]))
        encoded_labels = encoder.transform(["apple", "banana"])  # This will give you the encoded labels

        encoder.fit(["cherry"])  # Incrementally add "cherry"
        encoder.transform(["apple", "banana", "cherry"])  # Now it works, including "cherry"

        # Now you can use inverse_transform with the encoded labels
        print(encoder.classes_)
        original_labels = encoder.inverse_transform(encoded_labels)
        print(original_labels)  # This should print ['apple', 'banana']
    """

    def __init__(self):
        super().__init__()
        self.classes_ = np.array([])

    def _check_input(self, y):
        """
        Check and convert the input to a NumPy array if it is a list, tuple, pandas.Series, pandas.DataFrame, or torch.Tensor.

        Arguments:
            y (list, tuple, pd.Series, pd.DataFrame, torch.Tensor): The input labels.

        Returns:
            np.ndarray: The input labels converted to a NumPy array.
        """
        if isinstance(y, (list, tuple)):
            y = np.array(y)
        elif isinstance(y, pd.Series):
            y = y.values
        elif isinstance(y, torch.Tensor):
            y = y.numpy()
        return y

    def fit(self, y):
        """
        Fit the label encoder with an array of labels, incrementally adding new classes.

        Arguments:
            y (list, tuple, np.ndarray, pd.Series, pd.DataFrame, torch.Tensor): The input labels.

        Returns:
            IncrementalLabelEncoder: The instance itself.
        """
        y = self._check_input(y)
        new_unique_labels = np.unique(y)
        unique_labels = np.unique(
            np.concatenate((self.classes_, new_unique_labels))
        )
        self.classes_ = unique_labels
        return self

    def transform(self, y):
        """
        Transform labels to normalized encoding.

        Arguments:
            y (list, tuple, np.ndarray, pd.Series, pd.DataFrame, torch.Tensor): The input labels.

        Returns:
            np.ndarray: The encoded labels as a NumPy array.

        Raises:
            ValueError: If the input contains new labels that haven't been seen during `fit`.
        """

        y = self._check_input(y)
        diff = set(y) - set(self.classes_)
        if diff:
            raise ValueError(f"y contains new labels: {diff}")
        return super().transform(y)

    def inverse_transform(self, y):
        """
        Transform labels back to original encoding.

        Arguments:
            y (np.ndarray): The encoded labels as a NumPy array.

        Returns:
            np.ndarray: The original labels as a NumPy array.
        """

        return super().inverse_transform(y)


# # Obsolete warning for future compatibility
# class LabelEncoder(IncrementalLabelEncoder):
#     def __init__(self, *args, **kwargs):
#         """
#         Initialize the LabelEncoder with a deprecation warning.
#         """
#         warn(
#             "LabelEncoder is now obsolete; use IncrementalLabelEncoder instead.",
#             category=FutureWarning,
#         )
#         super().__init__(*args, **kwargs)


if __name__ == "__main__":
    # Example usage of IncrementalLabelEncoder
    le = LabelEncoder()
    le.fit(["A", "B"])
    print(le.classes_)

    le.fit(["C"])
    print(le.classes_)

    le.inverse_transform([0, 1, 2])

    le.fit(["X"])
    print(le.classes_)

    le.inverse_transform([3])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-04-05 12:14:08 (ywatanabe)"

from mngs.general import torch_fn


@torch_fn
def ensure_3d(x):
    if x.ndim == 1:  # assumes (seq_len,)
        x = x.unsqueeze(0).unsqueeze(0)
    elif x.ndim == 2:  # assumes (batch_siize, seq_len)
        x = x.unsqueeze(1)
    return x


# @torch_fn
# def unbias(x, dim=-1, fn="mean"):
#     if fn == "mean":
#         return x - x.mean(dim=dim, keepdims=True)
#     if fn == "min":
#         return x - x.min(dim=dim, keepdims=True)[0]



#!/usr/bin/env python3


from . import PARAMS, add_noise, filt, norm, reference, utils
from ._crop import crop
from ._demo_sig import demo_sig
from ._detect_ripples import _detect_ripples_preprocess, detect_ripples

# from ._ensure_3d import ensure_3d
from ._hilbert import hilbert

# from ._listen import listen
from ._misc import ensure_3d
from ._mne import get_eeg_pos
from ._modulation_index import modulation_index
from ._pac import pac
from ._psd import psd
from ._resample import resample
from ._transform import to_segments, to_sktime_df
from ._wavelet import wavelet



# Ranger deep learning optimizer - RAdam + Lookahead + Gradient Centralization, combined into one optimizer.

# https://github.com/lessw2020/Ranger-Deep-Learning-Optimizer
# and/or
# https://github.com/lessw2020/Best-Deep-Learning-Optimizers

# Ranger has been used to capture 12 records on the FastAI leaderboard.

# This version = 2020.9.4


# Credits:
# Gradient Centralization --> https://arxiv.org/abs/2004.01461v2 (a new optimization technique for DNNs), github:  https://github.com/Yonghongwei/Gradient-Centralization
# RAdam -->  https://github.com/LiyuanLucasLiu/RAdam
# Lookahead --> rewritten by lessw2020, but big thanks to Github @LonePatient and @RWightman for ideas from their code.
# Lookahead paper --> MZhang,G Hinton  https://arxiv.org/abs/1907.08610

# summary of changes:
# 9/4/20 - updated addcmul_ signature to avoid warning.  Integrates latest changes from GC developer (he did the work for this), and verified on performance on private dataset.
# 4/11/20 - add gradient centralization option.  Set new testing benchmark for accuracy with it, toggle with use_gc flag at init.
# full code integration with all updates at param level instead of group, moves slow weights into state dict (from generic weights),
# supports group learning rates (thanks @SHolderbach), fixes sporadic load from saved model issues.
# changes 8/31/19 - fix references to *self*.N_sma_threshold;
# changed eps to 1e-5 as better default than 1e-8.

import math
import torch
from torch.optim.optimizer import Optimizer, required


def centralized_gradient(x, use_gc=True, gc_conv_only=False):
    '''credit - https://github.com/Yonghongwei/Gradient-Centralization '''
    if use_gc:
        if gc_conv_only:
            if len(list(x.size())) > 3:
                x.add_(-x.mean(dim=tuple(range(1, len(list(x.size())))), keepdim=True))
        else:
            if len(list(x.size())) > 1:
                x.add_(-x.mean(dim=tuple(range(1, len(list(x.size())))), keepdim=True))
    return x


class Ranger(Optimizer):

    def __init__(self, params, lr=1e-3,                       # lr
                 alpha=0.5, k=6, N_sma_threshhold=5,           # Ranger options
                 betas=(.95, 0.999), eps=1e-5, weight_decay=0,  # Adam options
                 # Gradient centralization on or off, applied to conv layers only or conv + fc layers
                 use_gc=True, gc_conv_only=False, gc_loc=True
                 ):

        # parameter checks
        if not 0.0 <= alpha <= 1.0:
            raise ValueError(f'Invalid slow update rate: {alpha}')
        if not 1 <= k:
            raise ValueError(f'Invalid lookahead steps: {k}')
        if not lr > 0:
            raise ValueError(f'Invalid Learning Rate: {lr}')
        if not eps > 0:
            raise ValueError(f'Invalid eps: {eps}')

        # parameter comments:
        # beta1 (momentum) of .95 seems to work better than .90...
        # N_sma_threshold of 5 seems better in testing than 4.
        # In both cases, worth testing on your dataset (.90 vs .95, 4 vs 5) to make sure which works best for you.

        # prep defaults and init torch.optim base
        defaults = dict(lr=lr, alpha=alpha, k=k, step_counter=0, betas=betas,
                        N_sma_threshhold=N_sma_threshhold, eps=eps, weight_decay=weight_decay)
        super().__init__(params, defaults)

        # adjustable threshold
        self.N_sma_threshhold = N_sma_threshhold

        # look ahead params

        self.alpha = alpha
        self.k = k

        # radam buffer for state
        self.radam_buffer = [[None, None, None] for ind in range(10)]

        # gc on or off
        self.gc_loc = gc_loc
        self.use_gc = use_gc
        self.gc_conv_only = gc_conv_only
        # level of gradient centralization
        #self.gc_gradient_threshold = 3 if gc_conv_only else 1

        print(
            f"Ranger optimizer loaded. \nGradient Centralization usage = {self.use_gc}")
        if (self.use_gc and self.gc_conv_only == False):
            print(f"GC applied to both conv and fc layers")
        elif (self.use_gc and self.gc_conv_only == True):
            print(f"GC applied to conv layers only")

    def __setstate__(self, state):
        print("set state called")
        super(Ranger, self).__setstate__(state)

    def step(self, closure=None):
        loss = None
        # note - below is commented out b/c I have other work that passes back the loss as a float, and thus not a callable closure.
        # Uncomment if you need to use the actual closure...

        # if closure is not None:
        #loss = closure()

        # Evaluate averages and grad, update param tensors
        for group in self.param_groups:

            for p in group['params']:
                if p.grad is None:
                    continue
                grad = p.grad.data.float()

                if grad.is_sparse:
                    raise RuntimeError(
                        'Ranger optimizer does not support sparse gradients')

                p_data_fp32 = p.data.float()

                state = self.state[p]  # get state dict for this param

                if len(state) == 0:  # if first time to run...init dictionary with our desired entries
                    # if self.first_run_check==0:
                    # self.first_run_check=1
                    #print("Initializing slow buffer...should not see this at load from saved model!")
                    state['step'] = 0
                    state['exp_avg'] = torch.zeros_like(p_data_fp32)
                    state['exp_avg_sq'] = torch.zeros_like(p_data_fp32)

                    # look ahead weight storage now in state dict
                    state['slow_buffer'] = torch.empty_like(p.data)
                    state['slow_buffer'].copy_(p.data)

                else:
                    state['exp_avg'] = state['exp_avg'].type_as(p_data_fp32)
                    state['exp_avg_sq'] = state['exp_avg_sq'].type_as(
                        p_data_fp32)

                # begin computations
                exp_avg, exp_avg_sq = state['exp_avg'], state['exp_avg_sq']
                beta1, beta2 = group['betas']

                # GC operation for Conv layers and FC layers
                # if grad.dim() > self.gc_gradient_threshold:
                #    grad.add_(-grad.mean(dim=tuple(range(1, grad.dim())), keepdim=True))
                if self.gc_loc:
                    grad = centralized_gradient(grad, use_gc=self.use_gc, gc_conv_only=self.gc_conv_only)

                state['step'] += 1

                # compute variance mov avg
                exp_avg_sq.mul_(beta2).addcmul_(grad, grad, value=1 - beta2)

                # compute mean moving avg
                exp_avg.mul_(beta1).add_(grad, alpha=1 - beta1)

                buffered = self.radam_buffer[int(state['step'] % 10)]

                if state['step'] == buffered[0]:
                    N_sma, step_size = buffered[1], buffered[2]
                else:
                    buffered[0] = state['step']
                    beta2_t = beta2 ** state['step']
                    N_sma_max = 2 / (1 - beta2) - 1
                    N_sma = N_sma_max - 2 * \
                        state['step'] * beta2_t / (1 - beta2_t)
                    buffered[1] = N_sma
                    if N_sma > self.N_sma_threshhold:
                        step_size = math.sqrt((1 - beta2_t) * (N_sma - 4) / (N_sma_max - 4) * (
                            N_sma - 2) / N_sma * N_sma_max / (N_sma_max - 2)) / (1 - beta1 ** state['step'])
                    else:
                        step_size = 1.0 / (1 - beta1 ** state['step'])
                    buffered[2] = step_size

                # if group['weight_decay'] != 0:
                #    p_data_fp32.add_(-group['weight_decay']
                #                     * group['lr'], p_data_fp32)

                # apply lr
                if N_sma > self.N_sma_threshhold:
                    denom = exp_avg_sq.sqrt().add_(group['eps'])
                    G_grad = exp_avg / denom
                else:
                    G_grad = exp_avg

                if group['weight_decay'] != 0:
                    G_grad.add_(p_data_fp32, alpha=group['weight_decay'])
                # GC operation
                if self.gc_loc == False:
                    G_grad = centralized_gradient(G_grad, use_gc=self.use_gc, gc_conv_only=self.gc_conv_only)

                p_data_fp32.add_(G_grad, alpha=-step_size * group['lr'])
                p.data.copy_(p_data_fp32)

                # integrated look ahead...
                # we do it at the param level instead of group level
                if state['step'] % group['k'] == 0:
                    # get access to slow param tensor
                    slow_p = state['slow_buffer']
                    # (fast weights - slow weights) * alpha
                    slow_p.add_(p.data - slow_p, alpha=self.alpha)
                    # copy interpolated weights to RAdam param tensor
                    p.data.copy_(slow_p)

        return loss
